//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
//% NAMEDATE - Rename a file with today's date                              %
//% Written by James Greene, Schliengen, Germany                            %
//%-------------------------------------------------------------------------%
//% NAMEDATE                                                                %
//% Copyright (c) 1999-2009, 2014, 2015, 2016 James Greene                  %
//% NAMEDATE comes with NO WARRANTY to the extent permitted by law.         %
//% You may redistribute copies of NAMEDATE under the terms of the          %
//% GNU General Public License. For more information about these            %
//% matters, see the file named COPYING, or contact the author at           %
//% <james.greene@outlook.com>                                              %
//%                                                                         %
//% Please report any bugs to <james.greene@outlook.com>                    %
//%                                                                         %
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <conio.h>
#include <errno.h>
#include <time.h>
#include <io.h>
#include <direct.h>			// 24.04.2003 - Path support
#include <dirent.h>			// 19.06.2009 - opendir() support
#include <fcntl.h>
#include <locale.h>			// 7.09.2007 - locale support
#include <sys/types.h>
#include <sys/stat.h>
// #define WIN32_LEAN_AND_MEAN
#include <windows.h>


#define VER_NUMBER  "6.15"
#define VER_STRING  "Version " VER_NUMBER
#define VER_DATE    "08 Dec 2016"
#define COPR_TEXT	\
	"Copyright \xb8 1999, 2002-2009, 2014-2016 by James Greene. All rights reserved.\n" \
	"james.greene@outlook.com - www.informatics-consulting.de - ASP Member #06128\n\n"

#define	FMT_TEXT	 0
#define	FMT_DMY		 1
#define	FMT_YMD		 2
#define	FMT_USER	 4		// added 26.06.2003 (version 4.0)
#define	FMT_TEXTHM	 8		// added 27.09.2002
#define	FMT_DMYHM	 9		// added 27.09.2002
#define	FMT_YMDHM	10		// added 27.09.2002
#define	FMT_TIMEBIT	 8		// added 27.09.2002
#define FMT_USEREX	20		// added 02.02.2004 (version 5.0)

#define	FLG_CVTSPC	 1<<0		// added 13.10.2002
#define FLG_YES		 1<<1		// added 02.02.2004 (version 5.0)
#define FLG_FRONT	 1<<2		// added 25.03.2003 (version 2.1)
#define FLG_SYSTEM   1<<3		// added 24.04.2003 (version 3.0)
#define FLG_RENDIR   1<<4		// added 24.04.2003 (version 3.0)
#define FLG_FILETIME 1<<5		// added 11.05.2003 (version 3.1)
#define	FLG_OFFSET   1<<6		// added 11.05.2003 (version 3.1)
#define FLG_COPYFILE 1<<7	// added 12.06.2003 (version 3.2)
#define FLG_ASEXT	 1<<8	// added 30.06.2003 (version 4.0)
#define FLG_QUIET	 1<<9	// added 30.06.2003 (version 4.0)
// #define	FLG_OFFTIME  1<<10	// added  1.07.2003 (version 4.0) no longer needed as of v6.10 (combined with FLG_OFFSET)
#define	FLG_FORCE	 1<<11	// added  1.07.2003 (version 4.0)
#define	FLG_8DOT3	 1<<12	// added  23.10.2003 (version 4.1)
#define	FLG_CPY2DIR	 1<<13	// added  23.10.2003 (version 4.1)
#define	FLG_DYNAMIC  1<<14  // added 02.02.2004 (version 5.0)
#define FLG_LASTEXT	 1<<15	// added 24.05.2007 (version 5.2)
#define FLG_LOCALE	 1<<16	// added 7.09.2007 (version 6.0)
#define FLG_OEMCONV	 1<<17	// added 7.09.2007 (version 6.0)
#define FLG_OVERKILL 1<<18	// added 19.06.2009 (version 6.02)
#define FLG_SIMULATE 1<<19	// added 04.01.2015 (version 6.12)
#define FLG_MOVEFILE 1<<20  // added 05.01.2015 (version 6.13)
#define FLG_UTC      1<<21	// added 07.12.2016 (version 6.15)

typedef ULONG FLAGTYPE;		// added 7.09.2007 (version 6.0)

// static	int		nOffset=0;	// added 11.05.2003 (version 3.1)
// static  int		nOfftime=0;	// added 01.07.2003 (version 4.0)

#define	OFFSET_YEAR		0
#define	OFFSET_MONTH 	1
#define OFFSET_DAY		2
#define OFFSET_WEEK		3
#define	OFFSET_COUNT	4

#define	OFFTIME_HOUR	0
#define	OFFTIME_MIN		1
#define OFFTIME_SEC		2
#define	OFFTIME_COUNT	3

static	int		anOffset[OFFSET_COUNT] = {0, 0, 0, 0};			// added 24.06.09 (v6.10) - updated 08.12.16
static	int		anOfftime[OFFTIME_COUNT] = {0, 0, 0};	// added 24.06.09 (v6.10) - updated 08.12.16

static	UINT	MAX_ENTRIES = 256;

#define	MAXFORM	256					// it'll never be this long, but just to be safe...
static	char	strFormat[MAXFORM];	// date format
static	char	szTargetFolder[MAXFORM];	// target directory when /S command is used
static	char	strLocale[MAXFORM];			// country ID for locale

#define _countof(a) (sizeof(a)/sizeof(*(a)))	

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Help text
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static char *txtUsage[] = {
//  "----.----1----.----2----.----3----.----4----.----5----.----6----.----7----.----"
	" (**) /F = put the date before the original filename\n",
	" (**) /X = put the date as an extension to the original filename\n",
	"      /M  = use the file's modification date/time\n",
	"      /Z:\"[format]\"       = date format (see documentation)\n",
	"      /ZZ:\"[format]\"      = enhanced format (see documentation)\n",
	"      /E:\"[language]\"     = use [language] to for day and month names (ANSI)\n",
	"      /EE:\"[language]\"    = use [language] to for day and month names (ASCII)\n",
	"      /S:\"[path]\"         = rename or copy the files into directory [path]\n",
    "      /SM:\"[path]\"        = move and rename the files onto another drive\n",
	"      /O[Y|M|W|D]{+/-}[n] = add/subtract [n] Years/Months/Weeks/Days to date\n",
	"      /A[H|M|S]{+/-}[n]   = add/subtract [n] Hours/Minutes/Seconds to time\n",
	"      /B:[n]              = Set the buffer size to allow for up to n files.\n",
	"      /BD                 = Set the buffer size dynamically\n",
	"      /G  = use UTC instead of local timezone\n",
	"      /U  = convert any spaces in the filename to underscores\n",
	"      /D  = rename directories as well as files\n",
	"      /H  = rename system and hidden files\n",
	"      /C  = copy file instead of renaming it\n",
	"      /K  = force operation (Kill target file if it exists)\n",
	"      /KK = Kill target directory when used with /D (otherwise same as /K)\n",
	"      /L  = Use the LAST dot for the extension\n",
	"      /Y  = Automatically YES all confirmation checks.\n",
	"      /Q  = Quiet operation - no information messages\n",
	"      /P  = Pretend - Simulate the rename action (for debugging)\n",
	"      /V  = Display version information about NAMEDATE.\n",
	"      /?  = Display this help information.\n",
	"  (*) /8  = Use DOS 8.3 filenames (xxYYMMDD.ext)\n",
	"Legacy switches (Downward Compatibility)\n",
	"  (*) /N  = numeric form: Same as /Z:\"d-m-Y\"\n",
	"  (*) /R  = reverse numeric form: Same as /Z:\"Y-m-d\"\n",
	"  (*) /T  = add the time: Same as adding /Z:\"-H-M\"\n",
	"\n",
	"  (*) These options is not compatible with the /Z or /ZZ option.\n",
	" (**) These options are not compatible with the /ZZ option.\n\n"
};

#define	numUsage (sizeof(txtUsage) / sizeof (char *))

static	char *txtGNUDisclaimer[] = {
	"This program is free software: you can redistribute it and/or modify it under\n",
	"the terms of the GNU General Public License as published by the Free Software\n",
	"Foundation, either version 3 of the License, or (at your option) any later\n",
	"version.\n\n",
	"This program is distributed in the hope that it will be useful, but WITHOUT\n",
	"ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS\n",
	"FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.\n\n",
	"You should have received a copy of the GNU General Public License\n",
	"along with this program (COPYING.TXT), or see <http://www.gnu.org/licenses/>.\n"
};

#define	numDisclaim (sizeof(txtGNUDisclaimer) / sizeof (char *))

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static char *strcpy_s(char *pOut, unsigned uSize, const char *pIn) {
	char *rval = pOut;
	memset(pOut, 0, uSize);
	do {
		*pOut++ = *pIn++;
	} while (*pIn && uSize);
	return rval;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static void Version(void) {
	register int ri;
	
	fprintf(stderr, "NAMEDATE Version %s (%s at %.5s)\n", VER_NUMBER, VER_DATE, __TIME__);
	fputs(COPR_TEXT, stderr);
	for (ri = 0; ri < numDisclaim; ri++)
		fputs(txtGNUDisclaimer[ri], stderr);
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  void    Usage(char *exe) {
	int ri;
    fprintf(stderr, "Usage: %s [switches] filename\n\n", exe);
    fprintf(stderr, "where [switches] may include\n");
	for (ri = 0; ri < numUsage; ri++)
		fputs(txtUsage[ri], stderr);

	Version();
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// FormDate - Convert the user's specification into a strftime() format string

static	void	FormDate(const char *pForm, char *szFmt, BOOL fEx) {
	char *szF = szFmt;

	while (*pForm) {
		if (strchr("aAbBdHIjmMpSUwWyYzZ%", *pForm))	// all strftime flags EXCEPT %c, %x, and %X
			*szF++ = '%';
		else if (fEx && strchr("CFXc", *pForm))				// special handling for (F)ilename, e(X)t and (C)ounter
			*szF++ = '*';
		else if (*pForm == '\\')					// backslash escapes the next character
			++pForm;

		*szF++ = *pForm++;
	}
	*szF++ = '\0';
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	void	NeoDateTime(time_t now, const char *strFormat, char *szDate, const UINT cbDate, const FLAGTYPE nFlags) {
    struct tm *tmNow;
	
	tmNow = (nFlags & FLG_UTC) ? gmtime(&now) : localtime(&now); 

	strftime(szDate, cbDate, strFormat, tmNow);

	if (nFlags & FLG_OEMCONV) {
		char *sz = strdup(szDate);
		CharToOem(sz, szDate);
		free(sz);
	}
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	void	LegacyDate(const UINT nFormat) {
	char szForm[64];

	switch (nFormat & ~FMT_TIMEBIT) {
		case FMT_DMY :
			strcpy_s(szForm, _countof(szForm), "d-m-Y");
			break;
		case FMT_YMD :
			strcpy_s(szForm, _countof(szForm), "Y-m-d");
			break;
		case FMT_TEXT :
			strcpy_s(szForm, _countof(szForm), "d-b-Y");
	}

	if (nFormat & FMT_TIMEBIT)
		strcat(szForm, "-H-M");

	FormDate(szForm, strFormat, FALSE);
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	void	EightDotThreeDate(void) {
	FormDate("ymd", strFormat, FALSE);
}
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	int	copyfile(const char *szOldName, const char *szNewName, BOOL fForce) {
#ifdef _WIN32
	if(CopyFile(szOldName, szNewName, !fForce))
		return 0;
	else {
		errno = GetLastError();
		return -1;
	}
#else
	unsigned char bTemp[4096];
	int fin, fout, cnt, flg;


	fin = _open( szOldName, _O_RDONLY | _O_SEQUENTIAL | _O_BINARY);
	if( fin == -1 )
		return -1;

	flg = _O_WRONLY | _O_CREAT | _O_SEQUENTIAL | _O_BINARY;
	if (fForce)
		flg |= _O_TRUNC;

	fout = _open( szNewName, flg, _S_IREAD | _S_IWRITE );
	if( fout == -1 ) {
		_close(fin);
		return -1;
	}

	while ((cnt = _read(fin, bTemp, sizeof(bTemp))) > 0) {
		_write(fout, bTemp, cnt);
	}

	_close(fout);
	_close(fin);
	return 0;
#endif
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	int	exists(const char *strName) {
	int fin;

	fin = _open(strName, _O_RDONLY);
	if (fin == -1)
		return FALSE;
	_close(fin);
	return TRUE;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	int	direxists(const char *strName) {
	DIR *pdin;

	pdin = opendir(strName);
	if (pdin == 0)
		return FALSE;
	closedir(pdin);
	return TRUE;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// removedirectory - uses the Windows shell to remove a directory (including all contents!)
// returns 0 if successful, -1 upon failure with errno set.
// *** this function is, bluntly stated, fucking dangerous. Caveat Emptor. ***

static	int	removedirectory(const char *strRoot) {

	SHFILEOPSTRUCT shfo;
	TCHAR fqpn[MAX_PATH];
//	int ri;

	strcpy_s(fqpn, _countof(fqpn)-6, strRoot);
	strcat(fqpn, "\\");
	strcat(fqpn, "*.*\0\0");

	shfo.hwnd = GetDesktopWindow();
	shfo.wFunc = FO_DELETE;
	shfo.pFrom = fqpn;
	shfo.pTo = 0;
	shfo.fFlags = FOF_NOCONFIRMATION;
	shfo.lpszProgressTitle = "NameDate";

	(void) SHFileOperation(&shfo);

	return rmdir(strRoot);
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// NeoReplace - Handle /Z:"" replacements where the F, X and/or C are imbedded

static	void	NeoReplace(PSTR szNewName, UINT cbCnt, PCSTR szDate, PCSTR szGiven, PCSTR pExt, UINT uCount) {
	PSTR pT = szNewName;
	PCSTR pF = szDate;
	int nRest = (int)cbCnt;
	char chCvt[8];

	for (pF = szDate ; *pF ; pF++) {
		if (nRest <= 1)
			break;
		
		if (*pF == '*') {	// signal
			switch (*++pF) {
				case 'F' :	strcpy_s(pT, nRest, szGiven);			break;
				case 'X' :	strcpy_s(pT, nRest, pExt);				break;
				case 'C' :	sprintf(chCvt, "%04u", 1+uCount);	
							strcpy_s(pT, nRest, chCvt); 
							break;
				case 'c' :	sprintf(chCvt, "%02u", 1+uCount);
							strcpy_s(pT, nRest, chCvt); 
							break;
				default  :  *pT = '\0';						
			}
			while (*pT)
				++pT;
		} else
			*pT++ = *pF;
		nRest = (int)cbCnt - (int)(pT - szNewName);
	}

	*pT = '\0';
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	char	UserVerify(FLAGTYPE nFlags) {
	char ch;

	if (nFlags & (FLG_YES | FLG_SIMULATE)) {
		fputs("\n", stderr);
		fflush(stderr);
		return 'Y';
	}

	fputs("? (Yes, No, Quit): ", stderr);
	fflush(stderr);

	do {
		ch = _getch();
		ch = toupper(ch);
	} while (!strchr("YNQ", ch));

	switch (ch) {
		case 'Y' :
			fputs("YES\n", stderr);
			break;

		case 'N' :
			fputs("NO\n", stderr);
			break;

		case 'Q' :
			fputs("QUIT\n", stderr);
			break;
	}

	fflush(stderr);
	return ch;
}


//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	int	dorename(const char *szName, const time_t tmDate, const UINT nFormat, const FLAGTYPE nFlags, const UINT uCount) {
	char szGiven[MAX_PATH];
	char *pExt, *nExt, cSpc;
	char szNewName[MAX_PATH];
	char szDate[64];
	char chV;

	strcpy_s(szGiven, _countof(szGiven), szName);

	if ((pExt = strrchr(szGiven, '.')) != 0) {
		*pExt++ = '\0';
		if ((nFlags & FLG_LASTEXT) == 0) {		// added 24.05.2007 ...
			while ((nExt = strrchr(szGiven, '.')) != 0){
				*--pExt = '.';
				*nExt++ = '\0';
				pExt = nExt;
			}
		}
	} else {
		pExt = "";
	}

	cSpc = (nFlags & FLG_CVTSPC) ? '_' : ' ';

	// added 13.10.2002 - support for names without spaces...
	if (nFlags & FLG_CVTSPC) {
		char *cp;

		for (cp = szGiven; *cp; cp++) {
			if (*cp == ' ')
				*cp = '_';
		}
	}

	// +++ 11.05.2003 - now generate the date here
//	MakeDate(tmDate, nFormat, szDate);
	NeoDateTime(tmDate, strFormat, szDate, sizeof(szDate), nFlags);
	// --- 11.05.2003

	// +++ 02.02.2004 - new name generation (when using USEREX format)
	if (nFormat & FMT_USER) {
		if (nFormat == FMT_USEREX)
			NeoReplace(szNewName, _countof(szNewName), szDate, szGiven, pExt, uCount);
		else {
			char szTemp[MAX_PATH];	// added 23. Feb 2004 - /Z:"" wasn't working anymore ...

			if (nFlags & FLG_FRONT)
				sprintf(szTemp, "%s%c*F.*X", szDate, cSpc);		// 8.02.2005 - added the %c and cSpc, 23.11.2006 - corrected the order!
			else if (nFlags & FLG_ASEXT)
				sprintf(szTemp, "*F.%s.*X", szDate);
			else
				sprintf(szTemp, "*F%c%s.*X", cSpc, szDate);		// 8.02.2005 - added the %c and cSpc

			NeoReplace(szNewName, _countof(szNewName), szTemp, szGiven, pExt, uCount);
			// so, that should fix /Z again...
		}
	} else {
		if (nFlags & FLG_8DOT3)
			sprintf(szNewName, "%c%c%s.%s", 'A'+(uCount % 26), 'A'+(uCount / 26), szDate, pExt);
		else if (nFlags & FLG_FRONT)
			sprintf(szNewName, "%s%c%s.%s", szDate, cSpc, szGiven, pExt);
		else if (nFlags & FLG_ASEXT)
			sprintf(szNewName, "%s.%s.%s", szGiven, szDate, pExt);
		else
			sprintf(szNewName, "%s%c%s.%s", szGiven, cSpc, szDate, pExt);
	}

	if (nFlags & FLG_CPY2DIR) {
		char szTemp[MAX_PATH];
		strcpy_s(szTemp, _countof(szTemp)-strlen(szNewName), szTargetFolder);
		strcat(szTemp, szNewName);
		strcpy_s(szNewName, _countof(szNewName), szTemp);
	}

	if (((nFlags & (FLG_OVERKILL|FLG_RENDIR)) == (FLG_OVERKILL|FLG_RENDIR)) && direxists(szNewName)) {
		if (!(nFlags & FLG_QUIET)) {
			fprintf(stderr, "REMOVE DIRECTORY '%s'", szNewName);
			fflush(stderr);
		}

		if ((chV = UserVerify(nFlags)) == 'Y')
			if (removedirectory(szNewName))
				fprintf(stderr, "REMOVE DIRECTORY FAILED: %s\n", strerror(errno));

		if (chV == 'Q')
			return -2;
	} else {
		if ((nFlags & (FLG_FORCE|FLG_OVERKILL)) && exists(szNewName)) {
			if (!(nFlags & FLG_QUIET)) {
				fprintf(stderr, "DELETE '%s'", szNewName);
				fflush(stderr);
			}

			if ((chV = UserVerify(nFlags)) == 'Y')
				if (remove(szNewName))
					fprintf(stderr, "REMOVE FAILED: %s\n", strerror(errno));

			if (chV == 'Q')
				return -2;
		}
	}

	if (nFlags & (FLG_COPYFILE | FLG_MOVEFILE)) {
		if (!(nFlags & FLG_QUIET)) {
			fprintf(stderr, "%s '%s' TO '%s'", ((nFlags & FLG_MOVEFILE)?"MOVE":"COPY"), szName, szNewName);
			fflush(stderr);
		}

        if ((chV = UserVerify(nFlags)) == 'Y') {
			if (!(nFlags & FLG_SIMULATE) && copyfile(szName, szNewName, (nFlags & (FLG_FORCE|FLG_OVERKILL)))) {
				fprintf(stderr, "COPY FAILED: %s\n", strerror(errno));
				return errno;
			}
        
            if (nFlags & FLG_MOVEFILE) {    // added 2015-01-05
                if (!(nFlags & FLG_SIMULATE) && remove(szName)) {
                    fprintf(stderr, "REMOVE FAILED: %s\n", strerror(errno));
                    return errno;
                }
            }
        }
	} else {
		if (!(nFlags & FLG_QUIET)) {
			fprintf(stderr, "RENAME '%s' TO '%s'", szName, szNewName);
			fflush(stderr);
		}

		if ((chV = UserVerify(nFlags)) == 'Y')
			if (!(nFlags & FLG_SIMULATE) && rename(szName, szNewName)) {
				fprintf(stderr, "RENAME FAILED: %s\n", strerror(errno));
				return errno;
			}
	}

	return (chV == 'Q') ? -2 : 0;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	char *StrCopy(char *strTarget, const char *strSource, UINT nTarget) {
	strncpy(strTarget, strSource, nTarget-1);
	strTarget[nTarget-1] = 0;
	return strTarget;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static int	ChangeDirectory(const char *szName, int *pnCWDrv, char *pszCWDir, const UINT nCWDir, char *pszFile, const UINT nFile, UINT fSilent) {
	char *psl;		// path separator location
	UINT ri;		// index

	// get current working drive and directory
	*pnCWDrv = _getdrive();
	if (_getcwd(pszCWDir, nCWDir) == NULL)
		*pszCWDir = 0;

	// see if some separator (backslash or colon) is present
	if ((psl = strrchr(szName, '\\')) == NULL)
		psl = strchr(szName, ':');

	// if either a backslash or colon are present, then we need to change to that drive/path
	if (psl) {
		char szPath[MAX_PATH];
		int nDrv = *pnCWDrv;

		ri = (psl - szName);		// we should thrown an exception if ri > MAX_PATH here!
		StrCopy(szPath, szName, ri+1);
		if (pszFile)				// file can be null, so we need to check for this.
			StrCopy(pszFile, psl+1, nFile);

		if (szPath[1] == ':') {
			nDrv = 1 + (toupper(szPath[0]) - 'A');
#ifdef _DEBUG
			printf("... setting current drive to %c:\n", 64+nDrv);
#endif
			if (_chdrive(nDrv))
				perror("_chdrive failed");
		}

		if (strcmp(szPath+1, ":") == 0) 	// handle root directory situation
			strcat(szPath, "\\");

#ifdef _DEBUG
		printf("... setting current directory to %s\n", szPath);
#endif
		if (_chdir(szPath)) {
			if (errno == ENOENT) {		// directory not found
				if (pszFile == NULL) {	// we've been called by CreateTargetFolder, so let's try!
					if (_mkdir(szPath) == 0) {
						fprintf(stderr, "Directory %s created!\n", szPath);
						if (_chdir(szPath)) {
							fprintf(stderr, "Can't find directory %s: %s", szPath, strerror(errno));
							return 0;
						}
					} else {
						fprintf(stderr, "Error creating directory %s: %s", szPath, strerror(errno));
						return 0;
					}
				} else {
					fprintf(stderr, "Can't find directory %s: %s", szPath, strerror(errno));
					return 0;
				}
			} else {
				fprintf(stderr, "Can't find directory %s: %s", szPath, strerror(errno));
				return 0;
			}
		} else if (!fSilent)
			fprintf(stderr, "Directory set to %s\n", szPath);
	} else 	// no path separator found
		StrCopy(pszFile, szName, nFile);

	return 1;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
// Date / Time Offset handling
//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	int		isleapyr(const int yr) {
	return (((yr % 4) == 0) && ((yr % 100) || ((yr % 400) == 0))) ? 1 : 0;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// return the number of days in a given month of a given year

static	int		daysinmo(const int month, const int year) {
	int mdays[13] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};
	int adyr = year + 1900;

	if (month > 11)
		return 0;
	
	return mdays[month] + ((month == 1) ? isleapyr(adyr) : 0);
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// calculate the number of days between two month/year pairs

static	int		deltadays(const int month, const int year, const int dtmon, const int dtyr) {
	int nDays = 0;
	int yrmo = year*12 + month;
	int delta = dtyr*12 + dtmon;

#ifdef _DEBUG
	fprintf(stderr, "** deltadays: month=%d, year=%d, dtmon=%d, dtyr=%d\n", month, year, dtmon, dtyr);
#endif
	
	if (delta < 0) { // we're returning a negative value
		while (delta++) {
			nDays -= daysinmo(yrmo%12, yrmo/12);
			--yrmo;
#ifdef _DEBUG
	fprintf(stderr, "** deltadays: yrmo=%d, delta=%d, nDays=%d\n", yrmo, delta, nDays);
#endif
		}
	} else {
		while (delta--) {
			nDays += daysinmo(yrmo%12, yrmo/12);
			++yrmo;
#ifdef _DEBUG
	fprintf(stderr, "** deltadays: yrmo=%d, delta=%d, nDays=%d\n", yrmo, delta, nDays);
#endif
		}
	}
	
	return nDays;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

// removed a bunch of code here - the C runtime library will ensure that the month/year transitions
// are handled correctly. No need to bust my but doing the logic here.

static	time_t	timestamp(const time_t tOrg, const FLAGTYPE nFlags) {
	time_t tRet = tOrg;
	int dtDays = 0;

	if (nFlags & FLG_OFFSET) {		// offset specified?
			
		if (anOffset[OFFSET_MONTH] || anOffset[OFFSET_YEAR]) {	// we have a month or year offset
			struct tm *gmt;
			gmt = gmtime(&tOrg);
			dtDays = deltadays(gmt->tm_mon, gmt->tm_year, anOffset[OFFSET_MONTH], anOffset[OFFSET_YEAR]);
		}
	
		tRet += anOfftime[OFFTIME_SEC];				// seconds
		tRet += anOfftime[OFFTIME_MIN] * 60;		// minutes
		tRet += anOfftime[OFFTIME_HOUR] * 3600;		// hours
		
		dtDays += anOffset[OFFSET_DAY] + 7*anOffset[OFFSET_WEEK];

		tRet += dtDays * 86400;						// days
	}

	return tRet;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static  int namedate(const UINT nFormat, const FLAGTYPE nFlags, const char *szName, UINT *pnCount) {
	time_t now;
	long lFind;
	struct _finddata_t fi;
	UINT ri, uCnt=0;
	int nRV;
	int  nDrv;
	char szCWD[MAX_PATH];
//	char szPath[MAX_PATH];
	char szFile[MAX_PATH];
//	char *psl;
	char *bNames;
	time_t *bDate;
	int rv;
//	struct lconv *lc;

	// set the locale, if necessary
//	lc = localeconv();

	if (nFlags & FLG_LOCALE) {
		char *rsz = setlocale(LC_ALL, strLocale);
		fprintf(stderr, "Locale set to %s\n", rsz);
	}

	// get the current date and time
	now = time(NULL);

	// ++ 24.11.2006 ++ Path name support moved to ChangeDirectory ++
	ChangeDirectory(szName, &nDrv, szCWD, sizeof(szCWD), szFile, sizeof(szFile), 0);
	// -- 24.11.2006 -- End Path name --

	// if the wildcard option is selected, first collect all of the filenames,
	// then do the rename on each.

#ifdef _DEBUG
	{
		char szTemp[MAX_PATH];
		_getcwd(szTemp, sizeof(szTemp));
		printf("... CWD: %s\n", szTemp);
	}
#endif

	if (nFlags & FLG_DYNAMIC) {	// 2015-01-04 -- count how big the directory is...
		UINT nCount = 0;

		lFind = _findfirst(szFile, &fi);
		if (lFind == -1) {
			int err = errno;
			fprintf(stderr, "FAILED: %s: %s\n", szFile, strerror(err));
			return err;
		}
		
		do {
			if ((fi.attrib & _A_SUBDIR) && 
				(!(nFlags & FLG_RENDIR) || ((strcmp(fi.name, ".") == 0) || (strcmp(fi.name, "..") == 0))))
					continue;
		
			if ((fi.attrib & (_A_HIDDEN | _A_SYSTEM)) &&
				(!(nFlags & FLG_SYSTEM))) 
				continue;
			
			nCount++;
		} while (_findnext(lFind, &fi) == 0);
		
		MAX_ENTRIES = max(++nCount, MAX_ENTRIES);

		if (!(nFlags & FLG_QUIET)) {
			char szTemp[MAX_PATH];
			_getcwd(szTemp, sizeof(szTemp));
			fprintf(stderr, "DYNAMIC: Allocating space for %u files in '%s'\n", nCount, szTemp);
			fflush(stderr);
		}
		
	}
	
	bNames = (char *)calloc(MAX_ENTRIES, MAX_PATH);
	if (!bNames) {
		fprintf(stderr, "FAILED: Unable to allocate memory for %u files (%u Bytes total)\n", MAX_ENTRIES, MAX_ENTRIES*MAX_PATH);
		return -1;
	}
	
	bDate = (time_t *)calloc(MAX_ENTRIES, sizeof(time_t));
	if (!bDate) {
		fprintf(stderr, "FAILED: Unable to allocate memory for %u files (%u Bytes total)\n", MAX_ENTRIES, MAX_ENTRIES*sizeof(time_t));
		free(bNames);
		return -1;
	}

	lFind = _findfirst(szFile, &fi);
	if (lFind == -1) {
		int err = errno;
		fprintf(stderr, "FAILED: %s: %s\n", szFile, strerror(err));
		free(bNames);
		return err;
	}

	do {
		if (fi.attrib & _A_SUBDIR) {
			if (!(nFlags & FLG_RENDIR)) {
#ifdef _DEBUG
				printf("... skipping directory %s\n", fi.name);
#endif
				continue;
			}

			if ((strcmp(fi.name, ".") == 0) || (strcmp(fi.name, "..") == 0)) {
#ifdef _DEBUG
				printf("... skipping directory %s\n", fi.name);
#endif
				continue;
			}
		}

		if ((fi.attrib & (_A_HIDDEN | _A_SYSTEM)) &&
			(!(nFlags & FLG_SYSTEM))) {
#ifdef _DEBUG
			printf("... skipping system / hidden file %s\n", fi.name);
#endif
			continue;
		}

		strcpy_s(bNames+(uCnt * MAX_PATH), MAX_PATH, fi.name);
		bDate[uCnt++] = timestamp((nFlags & FLG_FILETIME) ? fi.time_write : now, nFlags);
	} while ((uCnt < MAX_ENTRIES) && (_findnext(lFind, &fi) == 0));

	for (ri=rv=0 ; (ri < uCnt) && (rv != -2) ; ri++) 
		rv = dorename(bNames+(ri * MAX_PATH), bDate[ri], nFormat, nFlags, ri);

	free(bDate);
	free(bNames);
	nRV = 0;

	// ++ 24.04.2003 ++ restore the current working directory
	_chdrive(nDrv);
	if (*szCWD)
		_chdir(szCWD);
	// -- 24.04.2003 --

	*pnCount += uCnt;

	return nRV;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

static	int		CreateTargetFolder(void) {
	char szRTD[MAX_PATH];
	char szCWD[MAX_PATH];
	int nCWDrv;
	UINT uvar;

	ChangeDirectory(szTargetFolder, &nCWDrv, szCWD, sizeof(szCWD), NULL, 0, 1);

	if (_getcwd(szRTD, sizeof(szRTD)) == NULL)
		*szRTD = 0;

	_chdrive(nCWDrv);
	if (*szCWD)
		_chdir(szCWD);

	StrCopy(szTargetFolder, szRTD, sizeof(szTargetFolder));

	uvar = strlen(szTargetFolder) - 1;
	if ((szTargetFolder[uvar] != '\\') && (szTargetFolder[uvar] != ':'))
		strcat(szTargetFolder, "\\");

	return 1;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

int main(int argc, char *argv[]) {
	char *ap, *exe, ch;
	UINT nFormat = FMT_TEXT;
	FLAGTYPE nFlags = 0;
//	BOOL flagTime = FALSE;
	char szOffset[128], *pOffset, chQ;
	UINT uvar;
	UINT nCount = 0;

	printf("NAMEDATE: Add a date to a given filename                         %14s\n", VER_STRING);
	fputs(COPR_TEXT, stdout);

	if (argc < 2) {
		Usage(*argv);
		return 1;
	}

	exe = *argv;

	while ((ap = *++argv) != 0) {
		if ((*ap == '/') || (*ap == '-')) {
			while (*++ap) {
				switch (toupper(*ap)) {
					case 'N' :	if ((nFormat & FMT_USER) == 0) {
									nFormat &= ~FMT_YMD ; nFormat |= FMT_DMY;
								} else
									fprintf(stderr, "-- Depreciated Parameter: /N not compatible with /Z\n");
								break;
					case 'R' :	if ((nFormat & FMT_USER) == 0) {
									nFormat &= ~FMT_DMY ; nFormat |= FMT_YMD;
								} else
									fprintf(stderr, "-- Depreciated Parameter: /R not compatible with /Z\n");
							   break;
					case 'T' :	if ((nFormat & FMT_USER) == 0)
									nFormat |= FMT_TIMEBIT;
								else
									fprintf(stderr, "-- Depreciated Parameter: /T not compatible with /Z\n");
								break;
					case 'F' :	if ((nFormat & FMT_USEREX) == 0)
//									nFormat |= (FLG_FRONT << 16);
									nFlags |= FLG_FRONT;
								else
									fprintf(stderr, "-- Depreciated Parameter: /F not compatible with /ZZ\n");
								break;
					case 'X' :	if ((nFormat & FMT_USEREX) == 0)
//									nFormat |= (FLG_ASEXT << 16);
									nFlags |= FLG_ASEXT;
								else
									fprintf(stderr, "-- Depreciated Parameter: /X not compatible with /ZZ\n");
								break;

					case 'M' : nFlags |= FLG_FILETIME;	break;
					case 'U' : nFlags |= FLG_CVTSPC;	break;
					case 'W' :                          break;
					case 'D' : nFlags |= FLG_RENDIR;	break;
					case 'H' : nFlags |= FLG_SYSTEM;	break;
					case 'C' : nFlags |= FLG_COPYFILE;	break;
					case 'Q' : nFlags |= FLG_QUIET;		break;
					case 'Y' : nFlags |= FLG_YES;		break;
					case 'L' : nFlags |= FLG_LASTEXT;	break;		// added 24.05.2007
					case 'P' : nFlags |= FLG_SIMULATE;	break;		// added 04.01.2015
					case 'G' : nFlags |= FLG_UTC;		break;		// specify UTC instead of localtime - added 07.12.2016

					case 'K' :
						ch = toupper(ap[1]);
						nFlags |= (ch == 'K') ? FLG_OVERKILL : FLG_FORCE;
						if (ch == 'K')
							++ap;
						break;

					case '8' :	if ((nFormat & FMT_USER) == 0)
//									nFormat |= (FLG_8DOT3 << 16);
									nFlags |= FLG_8DOT3;
								else
									fprintf(stderr, "-- Depreciated Parameter: /8 not compatible with /Z\n");
								break;

					case 'A' :	ch = toupper(*ap);
								chQ = toupper(*++ap);
								if (isdigit(chQ) || (chQ == '+') || (chQ == '-')) {
									fprintf(stderr, "-- Depreciated Parameter: /A[n]. Assuming /AM[n]\n");
									chQ = 'M';
									--ap;
								}
								while (*++ap == ' ')
									;
								pOffset = szOffset;
								while (isdigit(*ap) || (*ap == '+') || (*ap == '-'))
									*pOffset++ = *ap++;
								*pOffset = 0;
								--ap;
								switch (chQ) {
									case 'H' : 	uvar = OFFTIME_HOUR;	break;
									case 'M' : 	uvar = OFFTIME_MIN;		break;
									case 'S' :	uvar = OFFTIME_SEC;		break;
									default  : 	fprintf(stderr, "-- Invalid Parameter: /%c%c%s. Assuming /AM%d\n", ch, chQ, szOffset, atoi(szOffset));
												uvar = 1;
								}

								anOfftime[uvar] = atoi(szOffset);
								nFlags |= FLG_OFFSET;
								break;

					case 'O' :	ch = toupper(*ap);
								chQ = toupper(*++ap);
								if (isdigit(chQ) || (chQ == '+') || (chQ == '-')) {
									fprintf(stderr, "-- Depreciated Parameter: /O[n]. Assuming /OD[n]\n");
									chQ = 'D';
									--ap;
								}
								while (*++ap == ' ')
									;
								pOffset = szOffset;
								while (isdigit(*ap) || (*ap == '+') || (*ap == '-'))
									*pOffset++ = *ap++;
								*pOffset = 0;
								--ap;
								switch (chQ) {
									case 'Y' : 	uvar = OFFSET_YEAR;		break;
									case 'M' : 	uvar = OFFSET_MONTH;	break;
									case 'D' :	uvar = OFFSET_DAY;		break;
									case 'W' :	uvar = OFFSET_WEEK;		break;
									default  :	fprintf(stderr, "-- Invalid Parameter: /%c%c%s. Assuming /OM%d\n", ch, chQ, szOffset, atoi(szOffset));
												uvar = OFFSET_MONTH;
								}

								anOffset[uvar] = atoi(szOffset);
								nFlags |= FLG_OFFSET;
								break;


					case 'B' :		// set buffer size (v5.0)
						if (toupper(ap[1] == 'D')) {		// dynamic - added 2015-01-04
							nFlags |= FLG_DYNAMIC;
							++ap;
							break;
						}

						for (++ap; strchr(" :=", *ap); ap++)
							;

						pOffset = szOffset;
						while (isdigit(*ap))
							*pOffset++ = *ap++;
						*pOffset = 0;
						--ap;

						MAX_ENTRIES = atoi(szOffset);
						MAX_ENTRIES = max(MAX_ENTRIES, 256);
						break;

					case 'Z' :
						++ap;
//						nFormat = (HIWORD(nFormat) << 16) | ((toupper(*ap) == 'Z') ? FMT_USEREX : FMT_USER);
						nFormat = ((toupper(*ap) == 'Z') ? FMT_USEREX : FMT_USER);

						if ((toupper(*ap) == 'Z'))
							++ap;
						for (; strchr(" :=", *ap); ap++)
								   ;
						if (strchr("\"\'", *ap))
							chQ = *ap++;
						else
							chQ = 0;

						pOffset = szOffset;
						while (*ap && (*ap != chQ))
							*pOffset++ = *ap++;
						*pOffset = 0;
						FormDate(szOffset, strFormat, ((nFormat & FMT_USEREX) != FMT_USER));
						if (!*ap)
							--ap;
						break;

					case 'S' :	// /S="target path" - Files will be placed in this folder... added 23.10.2003
                        if (toupper(ap[1]) == 'M') {
                            nFlags |= FLG_MOVEFILE;
                            ++ap;
                        }
                        
						while (strchr(" \t:=", *++ap))
								   ;
						if (strchr("\"\'", *ap))
							chQ = *ap++;
						else
							chQ = 0;

						pOffset = szOffset;
						while (*ap && (*ap != chQ))
							*pOffset++ = *ap++;
						*pOffset = 0;

						strcpy_s(szTargetFolder, _countof(szTargetFolder), szOffset);
						uvar = (pOffset - szOffset);
						if (uvar > 1) {
							uvar--;
							if ((szTargetFolder[uvar] != '\\') && (szTargetFolder[uvar] != ':'))
								strcat(szTargetFolder, "\\");

//							nFormat |= (FLG_CPY2DIR << 16);
							nFlags |= FLG_CPY2DIR;
						}
						if (!*ap)
							--ap;
						break;

					case 'E' :										// added 7.09.2007
						++ap;
						nFlags |= FLG_LOCALE;
						if (toupper(*ap) == 'E') {
							nFlags |= FLG_OEMCONV;
							++ap;
						}

						for ( ; strchr(" :=", *ap); ap++)
							;
						if (strchr("\"\'", *ap))
							chQ = *ap++;
						else
							chQ = 0;

						pOffset = szOffset;
						while (*ap && (*ap != chQ))
							*pOffset++ = *ap++;
						*pOffset = 0;

						strcpy_s(strLocale, _countof(strLocale), szOffset);
						if (!*ap)
							--ap;
						break;
						
					case 'V' :	// added 24.06.2009
						Version();
						break;

					case '?' : Usage(exe); return 0;
					default  : fprintf(stderr, "-- Invalid Parameter: /%c ignored\n", *ap);
				}
			}
		} else {
			if (!(nFormat & FMT_USER))
				LegacyDate(nFormat);

			if (nFlags & FLG_8DOT3)	// override /U if /8 is specified
				EightDotThreeDate();

			if (nFlags & FLG_CPY2DIR)	// make the target directory a full path, and create the directory if needed.
				CreateTargetFolder();

		    namedate(nFormat, nFlags, ap, &nCount);
		}
	}

	return 0;
}

//%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
